import json
import requests

from datetime import datetime

from visonic.exceptions import *

class API(object):
    """ Class used for communication with the Visonic API """

    # Client configuration
    __app_type = 'com.visonic.powermaxapp'
    __user_agent = 'Dart/2.10 (dart:io)'
    __rest_version = '9.0'

    # API tokens
    __user_token = None
    __session_token = None

    # Use a session to reuse one TCP connection instead of creating a new
    # connection for every call to the API
    __session = None
    __timeout = 4

    def __init__(self, hostname, app_id):
        """ Class constructor initializes all URL variables. """

        # Set connection specific details
        self.__hostname = hostname
        self.__app_id = app_id

        self.render_urls()

        # Create a new session
        self.__session = requests.session()

    def render_urls(self):
        """ Configure the API endpoints. """
        self.__url_base = 'https://' + self.__hostname + '/rest_api/' + self.__rest_version
        self.__url_version = 'https://' + self.__hostname + '/rest_api/version'

        # API endpoints
        self.__url_access_grant             = self.__url_base + '/access/grant'
        self.__url_access_revoke            = self.__url_base + '/access/revoke'
        self.__url_activate_siren           = self.__url_base + '/activate_siren'
        self.__url_alarms                   = self.__url_base + '/alarms'
        self.__url_alerts                   = self.__url_base + '/alerts'
        self.__url_auth                     = self.__url_base + '/auth'
        self.__url_cameras                  = self.__url_base + '/cameras'
        self.__url_devices                  = self.__url_base + '/devices'
        self.__url_disable_siren            = self.__url_base + '/disable_siren'
        self.__url_events                   = self.__url_base + '/events'
        self.__url_feature_set              = self.__url_base + '/feature_set'
        self.__url_locations                = self.__url_base + '/locations'
        self.__url_notifications_email      = self.__url_base + '/notifications/email'
        self.__url_panel_login              = self.__url_base + '/panel/login'
        self.__url_panel_info               = self.__url_base + '/panel_info'
        self.__url_panel_add                = self.__url_base + '/panel/add'
        self.__url_panel_rename             = self.__url_base + '/panel/rename'
        self.__url_panel_unlink             = self.__url_base + '/panel/unlink'
        self.__url_panels                   = self.__url_base + '/panels'
        self.__url_password_reset           = self.__url_base + '/password/reset'
        self.__url_password_reset_complete  = self.__url_base + '/password/reset/complete'
        self.__url_process_status           = self.__url_base + '/process_status?process_tokens='
        self.__url_set_bypass_zone          = self.__url_base + '/set_bypass_zone'
        self.__url_set_name                 = self.__url_base + '/set_name'
        self.__url_set_state                = self.__url_base + '/set_state'
        self.__url_set_user_code            = self.__url_base + '/set_user_code'
        self.__url_smart_devices            = self.__url_base + '/smart_devices'
        self.__url_smart_devices_settings   = self.__url_base + '/smart_devices/settings'
        self.__url_status                   = self.__url_base + '/status'
        self.__url_troubles                 = self.__url_base + '/troubles'
        self.__url_users                    = self.__url_base + '/users'
        self.__url_wakeup_sms               = self.__url_base + '/wakeup_sms'

        # To be implemented

        # Will not be implemented
        self.__url_apptype                  = self.__url_base + '/apptype'
        self.__url_home_automation_devices  = self.__url_base + '/home_automation_devices'
        self.__url_make_video               = self.__url_base + '/make_video'

    def set_rest_version(self, version):
        """ Set which version to use when connection to the API. """
        self.__rest_version = version
        self.render_urls()

    def __raise_on_bad_request(self, error):
        """ Raise an exception when the API returns a bad request. """
        api = json.loads(error.decode('utf-8'))
        # print(api)

        if api['error'] == 10001: # BadRequestParams
            for pair in api['extras']:
                if pair['value'] == 'incorrect':
                    if pair['key'] == 'panel_serial':
                        raise PanelSerialIncorrectError()
                    elif pair['key'] == 'reset_password_code':
                        raise ResetPasswordCodeIncorrectError()
                if pair['value'] == 'required':
                    if pair['key'] == 'panel_serial':
                        raise PanelSerialRequiredError()
                    elif pair['key'] == 'email':
                        raise EmailRequiredError()
                    elif pair['key'] == 'password':
                        raise PasswordRequiredError()
                    elif pair['key'] == 'app_id':
                        raise AppIDRequiredError()
                    elif pair['key'] == 'user_code':
                        raise UserCodeRequiredError()
                    elif pair['key'] == 'new_password':
                        raise PasswordRequiredError()
                if pair['value'] == 'already_granted':
                    raise AlreadyGrantedError()
                if pair['value'] == 'already_linked':
                    raise AlreadyLinkedError()
                if pair['key'] == 'new_password':
                    raise NewPasswordStrengthError()
        elif api['error'] == 10004: # WrongCombination
            for pair in api['extras']:
                if pair['value'] == 'wrong_combination':
                    if pair['key'] == 'email' or pair['key'] == 'password':
                        raise WrongUsernameOrPasswordError()
                    if pair['key'] == 'panel_serial' or pair['key'] == 'master_user_code':
                        raise WrongPanelSerialOrMasterUserCodeError()
        elif api['error'] == 10021: # WrongUserCode
            raise UserCodeIncorrectError()
        elif api['error'] == 400 and api['error_reason_code'] == 'PanelNotConnected':
            raise PanelNotConnectedError()

        # Raise a generic error when the library has no
        # specific exception implemented yet.
        raise UndefinedBadRequestError(str(api))

    def __raise_on_forbidden(self, error):
        """ Raise an exception when the API returns a forbidden error. """
        api = json.loads(error.decode('utf-8'))
        # print(api)

        if api['error'] == 10010: # NotAllowed
            raise NotAllowedError()
        elif api['error'] == 10002: # UserAuthRequired
            raise UserAuthRequiredError()

        # Raise a generic error when the library has no
        # specific exception implemented yet.
        raise UndefinedForbiddenError(str(api))

    def __raise_on_unauthorized(self, error):
        """ Raise an exception when the API returns a unauthorized error. """
        api = json.loads(error.decode('utf-8'))
        # print(api)

        # Raise an exception when we are not authorized to access the endpoint
        raise UnauthorizedError(str(api))

    def __send_request(self, url, with_session_token=True, with_user_token=True, data_json=None, request_type='GET'):
        """ Send a GET or POST request to the server. Includes the Session-Token
        only if with_session_token is True. """

        # Prepare the headers to be sent
        headers = {
            'Host': self.__hostname,
            'Connection': 'keep-alive',
            'Accept': '*/*',
            'User-Agent': self.__user_agent,
            'Accept-Language': 'en-us',
            'Accept-Encoding': 'gzip',
        }

        # Only needed for POST requests
        if request_type == 'POST':
            headers['Content-Type'] = 'application/json'
            headers['Content-Length'] = str(len(data_json))

        # Include the session token in the header
        if with_session_token:
            headers['Session-Token'] = self.__session_token

        # Include the user authentication token in the header
        if with_user_token:
            headers['User-Token'] = self.__user_token

        # Perform the request and raise an exception
        # if the response is not OK (HTML 200)
        try:
            if request_type == 'GET':
                response = self.__session.get(url, headers=headers, timeout=self.__timeout)
            elif request_type == 'POST':
                response = self.__session.post(url, headers=headers, data=data_json, timeout=self.__timeout)
            response.raise_for_status()
        except requests.exceptions.ConnectTimeout:
            raise ConnectionTimeoutError(f"Connection to '{self.__hostname}' timed out after {str(self.__timeout)} seconds.")
            return None
        except requests.exceptions.HTTPError as e:
            api = json.loads(response.content.decode('utf-8'))
            if   '400 Client Error: Bad Request' in str(e):
                self.__raise_on_bad_request(response.content)
            elif '401 Client Error: Unauthorized' in str(e):
                self.__raise_on_unauthorized(response.content)
            elif '403 Client Error: Forbidden' in str(e):
                self.__raise_on_forbidden(response.content)
            elif '404 Client Error: Not Found' in str(e):
                raise NotFoundError()
            elif '420 Client Error:' in str(e):
                # TODO: {'error': 10020, 'error_message': 'Login temporary blocked', 'error_reason_code': 'LoginTemporaryBlocked', 'extras': [{'key': 'timeout', 'value': 44}]} // 44 = seconds to unblocked
                # print(api)
                raise LoginTemporaryBlockedError(f"Login is temporary blocked due to too many failed login attempts ({api['extras'][0]['value']} seconds remaining).")
            elif '440 Client Error: Session token not found' in str(e):
                raise SessionTokenError()
            elif '442 Client Error: Login attempts limit reached' in str(e):
                raise LoginAttemptsLimitReachedError('Login attempts limit reached.')
            elif '444 Client Error: Wrong user code' in str(e):
                raise InvalidUserCodeError('Authentication failed due to wrong user code.')
            else:
                print(api)
                raise

        # Check HTTP response code
        if response.status_code == requests.codes.ok:
            return json.loads(response.content.decode('utf-8'))
        else:
            return None

    ######################
    # Public API methods #
    ######################

    @property
    def session_token(self):
        """ Property to keep track of the session token. """
        return self.__session_token

    @property
    def hostname(self):
        """ Property to keep track of the API servers hostname. """
        return self.__hostname

    @property
    def user_token(self):
        """ Property to keep track of the user token beeing assigned during authentication. """
        return self.__user_token

    @property
    def app_id(self):
        """ Property to keep track of the user id (UUID) beeing used. """
        return self.__app_id

    def get_version_info(self):
        """ Find out which REST API versions are supported. """
        return self.__send_request(self.__url_version,
                                       with_session_token=False,
                                       with_user_token=False,
                                       request_type='GET')

    def authenticate(self, email, password):
        """ Try to authenticate and get a user auth token. """
        auth_info = {
            'email': email,
            'password': password,
            'app_id': self.__app_id,
        }

        auth_json = json.dumps(auth_info, separators=(',', ':'))
        res = self.__send_request(self.__url_auth,
                                       with_session_token=False,
                                       with_user_token=False,
                                       data_json=auth_json,
                                       request_type='POST')
        if res is not None:
            self.__user_token = res['user_token']
            return True
        else:
            return False

    def is_logged_in(self):
        """ Check if the session token is still valid. """
        try:
            self.get_status()
            return True
        except requests.HTTPError:
            return False

    def access_grant(self, user_id, email):
        """ Grant a user access to the alarm panel via the API. """
        user_data = {'user': user_id, 'email': email}
        user_json = json.dumps(user_data, separators=(',', ':'))
        return self.__send_request(self.__url_access_grant, data_json=user_json, request_type='POST')

    def access_revoke(self, user_id):
        """ Revoke access to the alarm panel via the API for a user. """
        user_data = {'user': user_id}
        user_json = json.dumps(user_data, separators=(',', ':'))
        return self.__send_request(self.__url_access_revoke, data_json=user_json, request_type='POST')

    def activate_siren(self):
        """ Activate the siren (sound the alarm). """
        siren_data = {}
        siren_json = json.dumps(siren_data, separators=(',', ':'))
        return self.__send_request(self.__url_activate_siren, data_json=siren_json, request_type='POST')

    def disable_siren(self, mode):
        """ Disable the siren (mute the alarm). """
        siren_data = {'mode': mode}
        siren_json = json.dumps(siren_data, separators=(',', ':'))
        return self.__send_request(self.__url_disable_siren, data_json=siren_json, request_type='POST')

    def get_alarms(self):
        """ Get the current alarms. """
        return self.__send_request(self.__url_alarms, request_type='GET')

    def get_alerts(self):
        """ Get the current alerts. """
        return self.__send_request(self.__url_alerts, request_type='GET')

    def get_cameras(self):
        """ Get the cameras in the system. """
        return self.__send_request(self.__url_cameras, request_type='GET')

    def get_devices(self):
        """ Get all device specific information. """
        return self.__send_request(self.__url_devices, request_type='GET')

    def get_email_notifications(self):
        """ Get settings for the email notifications. """
        return self.__send_request(self.__url_notifications_email, request_type='GET')

    def get_events(self):
        """ Get the alarm panel events. """
        return self.__send_request(self.__url_events, request_type='GET')

    def get_feature_set(self):
        """ Get the alarm panel feature set. """
        return self.__send_request(self.__url_feature_set, request_type='GET')

    def get_locations(self):
        """ Get all locations in the alarm system. """
        return self.__send_request(self.__url_locations, request_type='GET')

    def get_panel_info(self):
        """ The general panel information is only supported in version 4.0. """
        return self.__send_request(self.__url_panel_info, request_type='GET')

    def get_panels(self):
        """ Get a list of panels. """
        return self.__send_request(self.__url_panels, request_type='GET')

    def get_process_status(self, process_token):
        """ Get the current status of a process running on API server. """
        url = self.__url_process_status + process_token
        return self.__send_request(url, request_type='GET')

    def get_smart_devices(self):
        """ Get a list of smart devices. """
        return self.__send_request(self.__url_smart_devices, request_type='GET')

    def get_smart_devices_settings(self):
        """ Get a list of smart devices settings. """
        return self.__send_request(self.__url_smart_devices_settings, request_type='GET')

    def get_status(self):
        """ Get the current status of the alarm system. """
        return self.__send_request(self.__url_status, request_type='GET')

    def get_troubles(self):
        """ Get the current troubles. """
        return self.__send_request(self.__url_troubles, request_type='GET')

    def get_users(self):
        """ Get information about the active users.
        Note: Only master users can see the active_user_ids! """
        return self.__send_request(self.__url_users, request_type='GET')

    def get_wakeup_sms(self):
        """ Get the settings needed to wake up the alarm panel via SMS. """
        return self.__send_request(self.__url_wakeup_sms, request_type='GET')

    def panel_add(self, alias, panel_serial, access_proof, master_user_code):
        """ Add a new alarm panel to the user account. A master user code is required. """
        panel_data = {
            'alias': alias, 
            'panel_serial': panel_serial, 
            'access_proof': access_proof, 
            'master_user_code': master_user_code
        }
        panel_json = json.dumps(panel_data, separators=(',', ':'))
        return self.__send_request(self.__url_panel_add, data_json=panel_json, request_type='POST')

    def panel_login(self, panel_serial, user_code):
        """ Try to login to the alarm panel and get a session token. """
        login_info = {
            'user_code': user_code,
            'app_type': self.__app_type,
            'app_id': self.__app_id,
            'panel_serial': panel_serial
        }

        login_json = json.dumps(login_info, separators=(',', ':'))
        res = self.__send_request(self.__url_panel_login,
                                       with_session_token=False,
                                       data_json=login_json,
                                       request_type='POST')
        if res is not None:
            self.__session_token = res['session_token']
            return True
        else:
            return False

    def panel_rename(self, alias, panel_serial):
        """ Rename an alarm panel. """
        panel_data = {
            'panel_serial': panel_serial, 
            'alias': alias, 
        }
        panel_json = json.dumps(panel_data, separators=(',', ':'))
        return self.__send_request(self.__url_panel_rename, data_json=panel_json, request_type='POST')

    def panel_unlink(self, panel_serial, password, app_id):
        """ Unlink an alarm panel from the user account. """
        panel_data = {
            'panel_serial': panel_serial, 
            'password': password, 
            'app_id': app_id, 
        }
        panel_json = json.dumps(panel_data, separators=(',', ':'))
        return self.__send_request(self.__url_panel_unlink, data_json=panel_json, request_type='POST')

    def password_reset(self, email):
        """ Request a password reset email. An email will be sent to the email address provided. """
        reset_data = {'email': email}
        reset_json = json.dumps(reset_data, separators=(',', ':'))
        return self.__send_request(self.__url_password_reset, data_json=reset_json, request_type='POST')

    def password_reset_complete(self, reset_password_code, new_password):
        """ Complete the password reset request. """
        reset_data = {'reset_password_code': reset_password_code, 'new_password': new_password, 'app_id': self.__app_id}
        reset_json = json.dumps(reset_data, separators=(',', ':'))
        return self.__send_request(self.__url_password_reset_complete, data_json=reset_json, request_type='POST')

    def set_email_notifications(self, mode):
        """ Set settings for the email notifications. """
        notification_data = {'mode': mode}
        notification_json = json.dumps(notification_data, separators=(',', ':'))
        return self.__send_request(self.__url_notifications_email, data_json=notification_json, request_type='POST')

    def set_bypass_zone(self, zone, set_enabled):
        """ Enable or disable bypass mode for a zone. """
        bypass_data = {'zone': zone, 'set': set_enabled}
        bypass_json = json.dumps(bypass_data, separators=(',', ':'))
        return self.__send_request(self.__url_set_bypass_zone, data_json=bypass_json, request_type='POST')

    def set_name(self, object_class, id, name):
        """ Set the name of any type of object in the alarm system. """
        name_data = {'class': object_class, 'id': id, 'name': name}
        name_json = json.dumps(name_data, separators=(',', ':'))
        return self.__send_request(self.__url_set_name, data_json=name_json, request_type='POST')

    def set_user_code(self, user_code, user_id):
        """ Set the code of a user in the alarm system. """
        code_data = {'user_code': user_code, 'user_id': user_id}
        code_json = json.dumps(code_data, separators=(',', ':'))
        return self.__send_request(self.__url_set_user_code, data_json=code_json, request_type='POST')

    def arm_home(self, partition):
        """ Arm in Home mode. """
        arm_info = {'partition': partition, 'state': 'HOME'}
        arm_json = json.dumps(arm_info, separators=(',', ':'))
        return self.__send_request(self.__url_set_state, data_json=arm_json, request_type='POST')

    def arm_away(self, partition):
        """ Arm in Away mode. """
        arm_info = {'partition': partition, 'state': 'AWAY'}
        arm_json = json.dumps(arm_info, separators=(',', ':'))
        return self.__send_request(self.__url_set_state, data_json=arm_json, request_type='POST')

    def disarm(self, partition):
        """ Disarm the alarm system. """
        disarm_info = {'partition': partition, 'state': 'DISARM'}
        disarm_json = json.dumps(disarm_info, separators=(',', ':'))
        return self.__send_request(self.__url_set_state, data_json=disarm_json, request_type='POST')

    def send_get(self, url):
        """ Send a custom POST request. """
        return self.__send_request(url, request_type='GET')

    def send_post(self, url, data):
        """ Send a custom POST request. """
        data_json = json.dumps(data, separators=(',', ':'))
        return self.__send_request(url, data_json=data_json, request_type='POST')



import json
import requests

from dateutil import parser
from dateutil.relativedelta import *

from visonic.devices import *
from visonic.core import API
from visonic.exceptions import *
from visonic.classes import *


class Setup(object):
    """ Class definition of the main alarm system. """

    # API Connection
    __api = None

    def __init__(self, hostname, app_id, api_version='latest'):
        """ Initiate the connection to the REST API. """
        self.__api = API(hostname, app_id)
        self.set_rest_version(api_version)

    # System properties
    @property
    def api(self):
        """ Return the API for direct access. """
        return self.__api

    def access_grant(self, user_id, email):
        """ Grant a user access to the alarm panel via the API. """
        return self.__api.access_grant(user_id, email)

    def access_revoke(self, user_id):
        """ Revoke access to the alarm panel via the API for a user. """
        return self.__api.access_revoke(user_id)

    def activate_siren(self):
        """ Activate the siren (sound the alarm). """
        return self.__api.activate_siren()['process_token']

    def arm_home(self, partition=-1):
        """ Send Arm Home command to the alarm system. """
        return self.__api.arm_home(partition)['process_token']

    def arm_away(self, partition=-1):
        """ Send Arm Away command to the alarm system. """
        return self.__api.arm_away(partition)['process_token']

    def authenticate(self, email, password):
        """ Try to authenticate against the API with an email address and password. """
        return self.__api.authenticate(email, password)

    def connected(self):
        """ Check if the API server is connected to the alarm panel """
        return self.get_status().connected

    def disable_siren(self, mode='all'):
        """ Disable the siren (mute the alarm). """
        return self.__api.disable_siren(mode=mode)['process_token']

    def disarm(self, partition=-1):
        """ Send Disarm command to the alarm system. """
        return self.__api.disarm(partition)['process_token']

    def get_cameras(self):
        """ Fetch all the devices that are available. """
        camera_list = []

        cameras = self.__api.get_cameras()

        for camera in cameras:
            new_camera = Camera(
                location=camera['location'].capitalize(),
                partitions=camera['partitions'],
                preenroll=camera['preenroll'],
                preview_path=camera['preview_path'],
                status=camera['status'],
                timestamp=camera['timestamp'],
                zone=camera['zone'],
                zone_name=camera['zone_name'].capitalize(),
            )
            camera_list.append(new_camera)

        return camera_list

    def get_devices(self):
        """ Fetch all the devices that are available. """
        device_list = []

        devices = self.__api.get_devices()

        for device in devices:
            if device['subtype'] == 'CONTACT':
                contact_device = ContactDevice(
                    bypass=device['traits']['bypass']['enabled'] if 'bypass' in device['traits'] else False,
                    device_number=device['device_number'],
                    device_type=device['device_type'],
                    enrollment_id=device['enrollment_id'],
                    id=device['id'],
                    name=device['name'],
                    partitions=device['partitions'],
                    preenroll=device['preenroll'],
                    removable=device['removable'],
                    renamable=device['renamable'],
                    subtype=device['subtype'],
                    warnings=device['warnings'],
                    zone_type=device['zone_type'],
                    location=device['traits']['location']['name'].capitalize() if 'location' in device['traits'] else None,
                    soak=device['traits']['soak']['enabled'] if 'soak' in device['traits'] else False,
                )
                device_list.append(contact_device)
            elif device['subtype'] == 'MOTION_CAMERA':
                contact_device = CameraDevice(
                    bypass=device['traits']['bypass']['enabled'] if 'bypass' in device['traits'] else False,
                    device_number=device['device_number'],
                    device_type=device['device_type'],
                    enrollment_id=device['enrollment_id'],
                    id=device['id'],
                    name=device['name'],
                    partitions=device['partitions'],
                    preenroll=device['preenroll'],
                    removable=device['removable'],
                    renamable=device['renamable'],
                    subtype=device['subtype'],
                    warnings=device['warnings'],
                    zone_type=device['zone_type'],
                    location=device['traits']['location']['name'].capitalize() if 'location' in device['traits'] else None,
                    soak=device['traits']['soak']['enabled'] if 'soak' in device['traits'] else False,
                    vod=device['traits']['vod'] if 'vod' in device['traits'] else None,
                )
                device_list.append(contact_device)
            elif device['subtype'] == 'SMOKE':
                contact_device = SmokeDevice(
                    bypass=device['traits']['bypass']['enabled'] if 'bypass' in device['traits'] else False,
                    device_number=device['device_number'],
                    device_type=device['device_type'],
                    enrollment_id=device['enrollment_id'],
                    id=device['id'],
                    name=device['name'],
                    partitions=device['partitions'],
                    preenroll=device['preenroll'],
                    removable=device['removable'],
                    renamable=device['renamable'],
                    subtype=device['subtype'],
                    warnings=device['warnings'],
                    zone_type=device['zone_type'],
                    location=device['traits']['location']['name'].capitalize() if 'location' in device['traits'] else None,
                    soak=device['traits']['soak']['enabled'] if 'soak' in device['traits'] else False,
                )
                device_list.append(contact_device)
            elif device['subtype'] == 'BASIC_KEYFOB':
                contact_device = KeyFobDevice(
                    bypass=device['traits']['bypass']['enabled'] if 'bypass' in device['traits'] else False,
                    device_number=device['device_number'],
                    device_type=device['device_type'],
                    enrollment_id=device['enrollment_id'],
                    id=device['id'],
                    name=device['name'],
                    partitions=device['partitions'],
                    preenroll=device['preenroll'],
                    removable=device['removable'],
                    renamable=device['renamable'],
                    subtype=device['subtype'],
                    warnings=device['warnings'],
                    zone_type=device['zone_type'],
                    owner_id=device['traits']['owner']['id'] if 'owner' in device['traits'] else None,
                    owner_name=device['traits']['owner']['name'] if 'owner' in device['traits'] else None,
                )
                device_list.append(contact_device)
            elif device['device_type'] == 'GSM':
                contact_device = GSMDevice(
                    bypass=device['traits']['bypass']['enabled'] if 'bypass' in device['traits'] else False,
                    device_number=device['device_number'],
                    device_type=device['device_type'],
                    enrollment_id=device['enrollment_id'],
                    id=device['id'],
                    name=device['name'],
                    partitions=device['partitions'],
                    preenroll=device['preenroll'],
                    removable=device['removable'],
                    renamable=device['renamable'],
                    subtype=device['subtype'],
                    warnings=device['warnings'],
                    zone_type=device['zone_type'],
                    signal_level=device['traits']['signal_level']['level'] if 'signal_level' in device['traits'] else None,
                )
                device_list.append(contact_device)
            elif device['device_type'] == 'PGM':
                contact_device = PGMDevice(
                    bypass=device['traits']['bypass']['enabled'] if 'bypass' in device['traits'] else False,
                    device_number=device['device_number'],
                    device_type=device['device_type'],
                    enrollment_id=device['enrollment_id'],
                    id=device['id'],
                    name=device['name'],
                    partitions=device['partitions'],
                    preenroll=device['preenroll'],
                    removable=device['removable'],
                    renamable=device['renamable'],
                    subtype=device['subtype'],
                    warnings=device['warnings'],
                    zone_type=device['zone_type'],
                    parent_id=device['traits']['parent']['id'] if 'parent' in device['traits'] else None,
                    parent_port=device['traits']['parent']['port'] if 'parent' in device['traits'] else None,
                )
                device_list.append(contact_device)
            else:
                generic_device = GenericDevice(
                    bypass=device['traits']['bypass']['enabled'] if 'bypass' in device['traits'] else False,
                    device_number=device['device_number'],
                    device_type=device['device_type'],
                    enrollment_id=device['enrollment_id'],
                    id=device['id'],
                    name=device['name'],
                    partitions=device['partitions'],
                    preenroll=device['preenroll'],
                    removable=device['removable'],
                    renamable=device['renamable'],
                    subtype=device['subtype'],
                    warnings=device['warnings'],
                    zone_type=device['zone_type'],
                )
                device_list.append(generic_device)
        return device_list

    def get_events(self, timestamp_hour_offset=2):
        """ Get the last couple of events (60 events on my system). """
        event_list = []

        events = self.__api.get_events()

        for event in events:
            # Event timestamp
            dt = parser.parse(event['datetime'])
            dt = dt + relativedelta(hours=timestamp_hour_offset)
            timestamp = dt.strftime('%Y-%m-%d %H:%M:%S')
            event['datetime'] = timestamp

            new_event = Event(
                id=event['event'],
                type_id=event['type_id'],
                label=event['label'],
                description=event['description'],
                appointment=event['appointment'],
                datetime=event['datetime'],
                video=event['video'],
                device_type=event['device_type'],
                zone=event['zone'],
                partitions=event['partitions'],
                name=event['name'],
            )

            event_list.append(new_event)

        return event_list

    def get_feature_set(self):
        """ Fetch the locations associated with the alarm system. """
        feature_set = self.__api.get_feature_set()
        features = FeatureSet(
            events_enabled=feature_set['events']['is_enabled'],
            datetime_enabled=feature_set['datetime']['is_enabled'],
            partitions_enabled=feature_set['partitions']['is_enabled'],
            partitions_has_labels=feature_set['partitions']['is_labels_enabled'],
            partitions_max_count=feature_set['partitions']['max_partitions'],
            devices_enabled=feature_set['devices']['is_enabled'],
            sirens_can_enable=feature_set['sirens']['can_enable'],
            sirens_can_disable=feature_set['sirens']['can_disable'],
            home_automation_devices_enabled=feature_set['home_automation_devices']['is_enabled'],
            state_enabled=feature_set['state']['is_enabled'],
            state_can_set=feature_set['state']['can_set'],
            state_can_get=feature_set['state']['can_get'],
            faults_enabled=feature_set['faults']['is_enabled'],
            diagnostic_enabled=feature_set['diagnostic']['is_enabled'],
            wifi_enabled=feature_set['wifi']['is_enabled'],
        )
        return features

    def get_locations(self):
        """ Fetch the locations associated with the alarm system. """
        location_list = []
        for location in self.__api.get_locations():
            location_list.append(Location(location['hel_id'], location['name'].capitalize(), location['is_editable']))
        return location_list

    def get_panel_info(self):
        """ Fetch basic information about the alarm system. """
        gpi = self.__api.get_panel_info()

        return PanelInfo(gpi['current_user'], gpi['manufacturer'], gpi['model'], gpi['serial'])

    def get_panels(self):
        """ Fetch a list of panels associated with the user. """
        panel_list = []

        for panel in self.__api.get_panels():
            new_panel = Panel(
                panel_serial=panel['panel_serial'],
                alias=panel['alias'],
            )
            panel_list.append(new_panel)

        return panel_list

    def get_process_status(self, process_token):
        """ Fetch the status information associated with a process token. """
        process_list = []
        for process in self.__api.get_process_status(process_token):
            new_process = Process(
                token=process['token'],
                status=process['status'],
                message=process['message'],
                error=process['error'],
            )
            process_list.append(new_process)
        return process_list

    def get_rest_versions(self):
        """ Fetch the supported API versions. """
        return self.api.get_version_info()['rest_versions']

    def get_status(self):
        """ Fetch the current state of the alarm system. """

        status = self.__api.get_status()

        partition_list = []

        # Create the partitions
        for partition in status['partitions']:
            new_part = Partition(
                id=partition['id'], 
                state=partition['state'], 
                status=partition['status'], 
                ready=partition['ready'], 
                options=partition['options'],
            )
            partition_list.append(new_part)

        # Create the status
        new_status = Status(
            connected=status['connected'],
            bba_connected=status['connected_status']['bba']['is_connected'],
            bba_state=status['connected_status']['bba']['state'],
            gprs_connected=status['connected_status']['gprs']['is_connected'],
            gprs_state=status['connected_status']['gprs']['state'],
            discovery_completed=status['discovery']['completed'],
            discovery_stages=status['discovery']['stages'],
            discovery_in_queue=status['discovery']['in_queue'],
            discovery_triggered=status['discovery']['triggered'],
            partitions=partition_list,
            rssi_level=status['rssi']['level'],
            rssi_network=status['rssi']['network'],
        )

        return new_status

    def get_troubles(self):
        """ Fetch all the troubles that are available. """
        trouble_list = []
        for trouble in self.__api.get_troubles():
            new_trouble = Trouble(
                device_type=trouble['device_type'],
                location=trouble['location'],
                partitions=trouble['partitions'],
                trouble_type=trouble['trouble_type'],
                zone=trouble['zone'],
                zone_name=trouble['zone_name'],
                zone_type=trouble['zone_type'],
            )

            trouble_list.append(new_trouble)
        return trouble_list

    def get_users(self):
        """ Fetch a list of users in the alarm system. """
        users_info = self.__api.get_users()

        user_list = []

        for user in users_info['users']:
            user = User(user['id'], user['name'], user['email'], user['partitions'])
            user_list.append(user)

        return user_list

    def get_wakeup_sms(self):
        """ Fetch a list of users in the alarm system. """
        wakeup_sms = self.__api.get_wakeup_sms()
        sms = WakeupSMS(
            phone_number=wakeup_sms['phone'],
            message=wakeup_sms['sms'],
        )
        return sms

    def panel_add(self, alias, panel_serial, master_user_code, access_proof=None):
        """ Add a new alarm panel to the user account. A master user code is required. """
        return self.__api.panel_add(alias, panel_serial, access_proof, master_user_code)

    def panel_login(self, panel_serial, user_code):
        """ Establish a connection between the alarm panel and the API server. """
        return self.__api.panel_login(panel_serial, user_code)

    def panel_rename(self, alias, panel_serial):
        """ Rename an alarm panel. """
        return self.__api.panel_rename(alias, panel_serial)

    def panel_unlink(self, panel_serial, password, app_id):
        """ Unlink an alarm panel from the user account. """
        return self.__api.panel_unlink(panel_serial, password, app_id)

    def password_reset(self, email):
        """ Send a password reset link to the email address provided in the email argument. """
        return self.__api.password_reset(email)

    def password_reset_complete(self, reset_password_code, new_password):
        """ Complete the password reset by entering the reset code received in the email and a new password. """
        return self.__api.password_reset_complete(reset_password_code, new_password)['user_token']

    def set_bypass_zone(self, zone, set_enabled):
        """ Enabled or disable zone bypassing (for example, bypass a sensor to disable it). """
        return self.__api.set_bypass_zone(zone, set_enabled)['process_token']

    def set_name_user(self, user_id, name):
        """ Set the name of a user by user ID. """
        return self.__api.set_name('USER', user_id, name)['process_token']

    def set_rest_version(self, version='latest'):
        """ 
        Fetch the supported versions from the API server and automatically 
        configure the library to use the latest version supported by the server,
        unless overridden in the version parameter.
        """
        rest_versions = self.api.get_version_info()['rest_versions']
        rest_versions.sort(key=float)
        if version == "latest":
            self.__api.set_rest_version(rest_versions[-1])
        elif version in rest_versions:
            self.__api.set_rest_version(version)
        else:
            raise UnsupportedRestAPIVersionError(f'Rest API version {version} is not supported by server.')

    def set_user_code(self, user_id, user_code):
        """ Set the code of a user by user ID. """
        return self.__api.set_user_code(user_code, user_id)['process_token']




class Camera(object):
    """ Class definition of an event in the alarm system. """

    def __init__(self, location, partitions, preenroll, preview_path, status, timestamp, zone, zone_name):
        """ Set the private variable values on instantiation. """
        self.__location = location
        self.__partitions = partitions
        self.__preenroll = preenroll
        self.__preview_path = preview_path
        self.__status = status
        self.__timestamp = timestamp
        self.__zone = zone
        self.__zone_name = zone_name


    def __str__(self):
        """ Define how the print() method should print the object. """
        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """
        class_name   = type(self).__name__
        location     = f"location = {self.location}"
        partitions   = f"partitions = {self.partitions}"
        preenroll    = f"preenroll = {self.preenroll}"
        preview_path = f"preview_path = {self.preview_path}"
        status       = f"status = {self.status}"
        timestamp    = f"timestamp = {self.timestamp}"
        zone         = f"zone = {self.zone}"
        zone_name    = f"zone_name = {self.zone_name}"

        return f"{class_name}({location}, {partitions}, {preenroll}, {preview_path}, {status}, {timestamp}, {zone}, {zone_name})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'location': self.location,
            'partitions': self.partitions,
            'preenroll': self.preenroll,
            'preview_path': self.preview_path,
            'status': self.status,
            'timestamp': self.timestamp,
            'zone': self.zone,
            'zone_name': self.zone_name,
        }

    # Camera properties
    @property
    def location(self):
        return self.__location

    @property
    def partitions(self):
        return self.__partitions

    @property
    def preenroll(self):
        return self.__preenroll

    @property
    def preview_path(self):
        return self.__preview_path

    @property
    def status(self):
        return self.__status

    @property
    def timestamp(self):
        return self.__timestamp

    @property
    def zone(self):
        return self.__zone

    @property
    def zone_name(self):
        return self.__zone_name


class Event(object):
    """ Class definition of an event in the alarm system. """

    def __init__(self, id, type_id, label, description, appointment, datetime, video, device_type, zone, partitions, name):
        """ Set the private variable values on instantiation. """

        self.__id = id
        self.__type_id = type_id
        self.__label = label
        self.__description = description
        self.__appointment = appointment
        self.__datetime = datetime
        self.__video = video
        self.__device_type = device_type
        self.__zone = zone
        self.__partitions = partitions
        self.__name = name


    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name  = type(self).__name__
        id          = f"id = {self.id}"
        type_id     = f"type_id = {self.type_id}"
        label       = f"label = '{self.label}'"
        description = f"description = '{self.description}'"
        appointment = f"appointment = '{self.appointment}'"
        datetime    = f"datetime = '{self.datetime}'"
        video       = f"video = {self.video}"
        device_type = f"device_type = '{self.device_type}'"
        zone        = f"zone = {self.zone}"
        partitions  = f"partitions = {str(self.partitions)}"
        name        = f"name = {str(self.name)}"

        return f"{class_name}({id}, {type_id}, {label}, {description}, {appointment}, \
            {datetime}, {video}, {device_type}, {zone}, {partitions}, {name})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'id': self.id,
            'type_id': self.type_id,
            'label': self.label,
            'description': self.description,
            'appointment': self.appointment,
            'datetime': self.datetime,
            'video': self.video,
            'device_type': self.device_type,
            'zone': self.zone,
            'partitions': self.partitions,
            'name': self.name,
        }

   # Event properties
    @property
    def id(self):
        """ User ID. """
        return self.__id

    @property
    def type_id(self):
        """ Event type ID. """
        return self.__type_id

    @property
    def label(self):
        """ Event label. """
        return self.__label

    @property
    def description(self):
        """ Event description. """
        return self.__description

    @property
    def appointment(self):
        """ Event appointment. """
        return self.__appointment

    @property
    def datetime(self):
        """ Event datetime. """
        return self.__datetime

    @property
    def video(self):
        """ Event has video. """
        return self.__video

    @property
    def device_type(self):
        """ Event device type. """
        return self.__device_type

    @property
    def zone(self):
        """ Event zone. """
        return self.__zone

    @property
    def partitions(self):
        """ Event partitions. """
        return self.__partitions

    @property
    def name(self):
        """ Event name. """
        return self.__name


class FeatureSet(object):
    """ Class definition of an event in the alarm system. """

    def __init__(self, events_enabled, datetime_enabled, partitions_enabled, partitions_has_labels, partitions_max_count, devices_enabled, sirens_can_enable, sirens_can_disable, home_automation_devices_enabled, state_enabled, state_can_set, state_can_get, faults_enabled, diagnostic_enabled, wifi_enabled):
        """ Set the private variable values on instantiation. """

        self.__events_enabled = events_enabled
        self.__datetime_enabled = datetime_enabled
        self.__partitions_enabled = partitions_enabled
        self.__partitions_has_labels = partitions_has_labels
        self.__partitions_max_count = partitions_max_count
        self.__devices_enabled = devices_enabled
        self.__sirens_can_enable = sirens_can_enable
        self.__sirens_can_disable = sirens_can_disable
        self.__home_automation_devices_enabled = home_automation_devices_enabled
        self.__state_enabled = state_enabled
        self.__state_can_set = state_can_set
        self.__state_can_get = state_can_get
        self.__faults_enabled = faults_enabled
        self.__diagnostic_enabled = diagnostic_enabled
        self.__wifi_enabled = wifi_enabled


    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name  = type(self).__name__
        events_enabled = f"events_enabled = {self.events_enabled}"
        datetime_enabled = f"datetime_enabled = {self.datetime_enabled}"
        partitions_enabled = f"partitions_enabled = {self.partitions_enabled}"
        partitions_has_labels = f"partitions_has_labels = {self.partitions_has_labels}"
        partitions_max_count = f"partitions_max_count = {self.partitions_max_count}"
        devices_enabled = f"devices_enabled = {self.devices_enabled}"
        sirens_can_enable = f"sirens_can_enable = {self.sirens_can_enable}"
        sirens_can_disable = f"sirens_can_disable = {self.sirens_can_disable}"
        home_automation_devices_enabled = f"home_automation_devices_enabled = {self.home_automation_devices_enabled}"
        state_enabled = f"state_enabled = {self.state_enabled}"
        state_can_set = f"state_can_set = {self.state_can_set}"
        state_can_get = f"state_can_get = {self.state_can_get}"
        faults_enabled = f"faults_enabled = {self.faults_enabled}"
        diagnostic_enabled = f"diagnostic_enabled = {self.diagnostic_enabled}"
        wifi_enabled = f"wifi_enabled = {self.wifi_enabled}"

        return f"{class_name}({events_enabled}, {datetime_enabled}, {partitions_enabled}, {partitions_has_labels}, {partitions_max_count}, {devices_enabled}, {sirens_can_enable}, {sirens_can_disable}, {home_automation_devices_enabled}, {state_enabled}, {state_can_set}, {state_can_get}, {faults_enabled}, {diagnostic_enabled}, {wifi_enabled})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'events_enabled': self.events_enabled,
            'datetime_enabled': self.datetime_enabled,
            'partitions_enabled': self.partitions_enabled,
            'partitions_has_labels': self.partitions_has_labels,
            'partitions_max_count': self.partitions_max_count,
            'devices_enabled': self.devices_enabled,
            'sirens_can_enable': self.sirens_can_enable,
            'sirens_can_disable': self.sirens_can_disable,
            'home_automation_devices_enabled': self.home_automation_devices_enabled,
            'state_enabled': self.state_enabled,
            'state_can_set': self.state_can_set,
            'state_can_get': self.state_can_get,
            'faults_enabled': self.faults_enabled,
            'diagnostic_enabled': self.diagnostic_enabled,
            'wifi_enabled': self.wifi_enabled,
        }

   # Event properties
    @property
    def events_enabled(self):
        return self.__events_enabled

    @property
    def datetime_enabled(self):
        return self.__datetime_enabled

    @property
    def partitions_enabled(self):
        return self.__partitions_enabled

    @property
    def partitions_has_labels(self):
        return self.__partitions_has_labels

    @property
    def partitions_max_count(self):
        return self.__partitions_max_count

    @property
    def devices_enabled(self):
        return self.__devices_enabled

    @property
    def sirens_can_enable(self):
        return self.__sirens_can_enable

    @property
    def sirens_can_disable(self):
        return self.__sirens_can_disable

    @property
    def home_automation_devices_enabled(self):
        return self.__home_automation_devices_enabled

    @property
    def state_enabled(self):
        return self.__state_enabled

    @property
    def state_can_set(self):
        return self.__state_can_set

    @property
    def state_can_get(self):
        return self.__state_can_get

    @property
    def faults_enabled(self):
        return self.__faults_enabled

    @property
    def diagnostic_enabled(self):
        return self.__diagnostic_enabled

    @property
    def wifi_enabled(self):
        return self.__wifi_enabled


class Location(object):
    """ Class definition of a location in the alarm system. """

    def __init__(self, id, name, is_editable):
        """ Set the private variable values on instantiation. """

        self.__id = id
        self.__name = name
        self.__is_editable = is_editable

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """
        
        class_name   = type(self).__name__
        id           = f"id = {self.id}"
        name         = f"name = '{self.name}'"
        is_editable  = f"is_editable = {self.is_editable}"

        return f"{class_name}({id}, {name}, {is_editable})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'id': self.id,
            'name': self.name,
            'is_editable': self.is_editable,
        }

   # Location properties
    @property
    def id(self):
        """ Device ID. """
        return self.__id

    @property
    def name(self):
        """ Device name. """
        return self.__name

    @property
    def is_editable(self):
        """ Device is editable. """
        return self.__is_editable


class PanelInfo(object):
    """ Class definition of the general alarm system information. """

    def __init__(self, current_user, manufacturer, model, serial):
        """ Set the private variable values on instantiation. """

        self.__current_user = current_user
        self.__manufacturer = manufacturer
        self.__model = model
        self.__serial = serial

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name             = type(self).__name__
        current_user           = f"current_user = '{self.current_user}'"
        manufacturer           = f"manufacturer = '{self.manufacturer}'"
        model                  = f"model = '{self.model}'"
        serial                 = f"serial = {self.serial}"

        return f"{class_name}({current_user}, {manufacturer}, {model}, {serial})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'current_user': self.current_user,
            'manufacturer': self.manufacturer,
            'model': self.model,
            'serial': self.serial,
        }

    # PanelInfo properties
    @property
    def current_user(self):
        return self.__current_user

    @property
    def manufacturer(self):
        return self.__manufacturer

    @property
    def model(self):
        """ Model name. """
        return self.__model

    @property
    def serial(self):
        return self.__serial


class Panel(object):
    """ Class definition of the general alarm system information. """

    def __init__(self, panel_serial, alias):
        """ Set the private variable values on instantiation. """

        self.__panel_serial = panel_serial
        self.__alias = alias

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name   = type(self).__name__
        panel_serial = f"panel_serial = '{self.panel_serial}'"
        alias        = f"alias = '{self.alias}'"

        return f"{class_name}({panel_serial}, {alias})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'panel_serial': self.panel_serial,
            'alias': self.alias,
        }

    # Panel properties
    @property
    def panel_serial(self):
        return self.__panel_serial

    @property
    def alias(self):
        return self.__alias


class Partition(object):
    """ Class definition of a partition in the alarm system. """

    def __init__(self, id, state, status, ready, options):
        """ Set the private variable values on instantiation. """

        self.__id = id
        self.__state = state
        self.__status = status
        self.__ready = ready
        self.__options = options

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name = type(self).__name__
        id         = f"id = {self.id}"
        state      = f"state = '{self.state}'"
        status     = f"status = '{self.status}'"
        ready      = f"ready = {self.ready}"
        options    = f"options = {self.options}"

        return f"{class_name}({id}, {state}, {status}, {ready}, {options})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'id': self.id,
            'state': self.state,
            'status': self.status,
            'ready': self.ready,
            'options': self.options,
        }

    # Partition properties
    @property
    def id(self):
        return self.__id

    @property
    def state(self):
        return self.__state

    @property
    def status(self):
        return self.__status

    @property
    def ready(self):
        return self.__ready

    @property
    def options(self):
        return self.__options


class Process(object):
    """ Class definition of a process in the alarm system. """

    def __init__(self, token, status, message, error):
        """ Set the private variable values on instantiation. """

        self.__token = token
        self.__status = status
        self.__message = message
        self.__error = error

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name = type(self).__name__
        token   = f"token = '{self.token}'"
        status  = f"status = '{self.status}'"
        message = f"message = '{self.message}'"
        error   = f"error = '{self.error}'"

        return f"{class_name}({token}, {status}, {message}, {error})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'token': self.token,
            'status': self.status,
            'message': self.message,
            'error': self.error,
        }

    # Partition properties
    @property
    def token(self):
        return self.__token

    @property
    def status(self):
        return self.__status

    @property
    def message(self):
        return self.__message

    @property
    def error(self):
        return self.__error


class Status(object):
    """ Class definition representing the status of the alarm system. """

    def __init__(self, connected, bba_connected, bba_state, gprs_connected, gprs_state, discovery_completed, discovery_stages, discovery_in_queue, discovery_triggered, partitions, rssi_level, rssi_network):
        """ Set the private variable values on instantiation. """
        self.__connected = connected
        self.__bba_connected = bba_connected
        self.__bba_state = bba_state
        self.__gprs_connected = gprs_connected
        self.__gprs_state = gprs_state
        self.__discovery_completed = discovery_completed
        self.__discovery_stages = discovery_stages
        self.__discovery_in_queue = discovery_in_queue
        self.__discovery_triggered = discovery_triggered
        self.__partitions = partitions
        self.__rssi_level = rssi_level
        self.__rssi_network = rssi_network

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name = type(self).__name__
        connected = f"connected = {self.connected}"
        bba_connected = f"bba_connected = {self.bba_connected}"
        bba_state = f"bba_state = {self.bba_state}"
        gprs_connected = f"gprs_connected = {self.gprs_connected}"
        gprs_state = f"gprs_state = {self.gprs_state}"
        discovery_completed = f"discovery_completed = {self.discovery_completed}"
        discovery_stages = f"discovery_stages = {self.discovery_stages}"
        discovery_in_queue = f"discovery_in_queue = {self.discovery_in_queue}"
        discovery_triggered = f"discovery_triggered = {self.discovery_triggered}"
        partitions = f"partitions = [{str(self.partitions)}]"
        rssi_level = f"rssi_level = {self.rssi_level}"
        rssi_network = f"rssi_network = {self.rssi_network}"

        return f"{class_name}({connected}, {bba_connected}, {bba_state}, {gprs_connected}, \
                 {gprs_state}, {discovery_completed}, {discovery_stages}, {discovery_in_queue}, \
                 {discovery_triggered}, {partitions}, {rssi_level}, {rssi_network})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'connected': self.connected,
            'bba_connected': self.bba_connected,
            'bba_state': self.bba_state,
            'gprs_connected': self.gprs_connected,
            'gprs_state': self.gprs_state,
            'discovery_completed': self.discovery_completed,
            'discovery_stages': self.discovery_stages,
            'discovery_in_queue': self.discovery_in_queue,
            'discovery_triggered': self.discovery_triggered,
            'partitions': self.partitions,
            'rssi_level': self.rssi_level,
            'rssi_network': self.rssi_network,
        }

    # Status properties
    @property
    def connected(self):
        return self.__connected

    @property
    def bba_connected(self):
        return self.__bba_connected

    @property
    def bba_state(self):
        return self.__bba_state

    @property
    def gprs_connected(self):
        return self.__gprs_connected

    @property
    def gprs_state(self):
        return self.__gprs_state

    @property
    def discovery_completed(self):
        return self.__discovery_completed

    @property
    def discovery_stages(self):
        return self.__discovery_stages

    @property
    def discovery_in_queue(self):
        return self.__discovery_in_queue

    @property
    def discovery_triggered(self):
        return self.__discovery_triggered

    @property
    def partitions(self):
        return self.__partitions

    @property
    def rssi_level(self):
        return self.__rssi_level

    @property
    def rssi_network(self):
        return self.__rssi_network


class Trouble(object):
    """ Class definition of a trouble in the alarm system. """

    def __init__(self, device_type, location, partitions, trouble_type, zone, zone_name, zone_type):
        """ Set the private variable values on instantiation. """

        self.__device_type = device_type
        self.__location = location
        self.__partitions = partitions
        self.__trouble_type = trouble_type
        self.__zone = zone
        self.__zone_name = zone_name
        self.__zone_type = zone_type

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name   = type(self).__name__
        device_type  = f"device_type = '{self.device_type}'"
        location     = f"location = '{self.location}'"
        partitions   = f"partitions = {str(self.partitions)}"
        trouble_type = f"trouble_type = '{self.trouble_type}'"
        zone         = f"zone = {self.zone}"
        zone_name    = f"zone_name = '{self.zone_name}'"
        zone_type    = f"zone_type = '{self.zone_type}'"

        return f"{class_name}({device_type}, {location}, {partitions}, \
            {trouble_type}, {zone}, {zone_name}, {zone_type})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'device_type': self.device_type,
            'location': self.location,
            'partitions': self.partitions,
            'trouble_type': self.trouble_type,
            'zone': self.zone,
            'zone_name': self.zone_name,
            'zone_type': self.zone_type,
        }

    # Trouble properties
    @property
    def device_type(self):
        """ Device type. """
        return self.__device_type

    @property
    def location(self):
        """ Location. """
        return self.__location

    @property
    def partitions(self):
        """ Partitions. """
        return self.__partitions

    @property
    def trouble_type(self):
        """ Trouble type. """
        return self.__trouble_type

    @property
    def zone(self):
        """ Zone ID. """
        return self.__zone

    @property
    def zone_name(self):
        """ Zone type. """
        return self.__zone_name

    @property
    def zone_type(self):
        """ Zone type. """
        return self.__zone_type


class User(object):
    """ Class definition of a user in the alarm system. """

    def __init__(self, id, name, email, partitions):
        """ Set the private variable values on instantiation. """

        self.__id = id
        self.__name = name
        self.__email = email
        self.__partitions = partitions

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """

        class_name  = type(self).__name__
        id          = f"id = {self.id}"
        name        = f"name = '{self.name}'"
        email       = f"email = '{self.name}'"
        partitions  = f"partitions = {self.is_active}"

        return f"{class_name}({id}, {name}, {email}, {partitions})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'id': self.id,
            'name': self.name,
            'email': self.email,
            'partitions': self.partitions,
        }

   # User properties
    @property
    def id(self):
        """ User ID. """
        return self.__id

    @property
    def name(self):
        """ User name. """
        return self.__name

    @property
    def email(self):
        """ User email. """
        return self.__email

    @property
    def partitions(self):
        """ Device is active. """
        return self.__partitions


class WakeupSMS(object):
    """ Class definition of a wakeup SMS in the alarm system. """

    def __init__(self, phone_number, message):
        """ Set the private variable values on instantiation. """
        self.__phone_number = phone_number
        self.__message = message

    def __str__(self):
        """ Define how the print() method should print the object. """
        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented on output to console. """
        class_name   = type(self).__name__
        phone_number = f"phone_number = {self.phone_number}"
        message      = f"message = {self.message}"

        return f"{class_name}({phone_number}, {message})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'phone_number': self.phone_number,
            'message': self.message,
        }

   # Wakeup SMS properties
    @property
    def phone_number(self):
        return self.__phone_number

    @property
    def message(self):
        return self.__message




class Device(object):
    """ Base class definition of a device in the alarm system. """

    # Property variables
    __bypass = None
    __device_number = None
    __device_type = None
    __enrollment_id = None
    __id = None
    __name = None
    __partitions = None
    __preenroll = None
    __removable = None
    __renamable = None
    __subtype = None
    __warnings = None
    __zone_type = None

    def __init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type):
        """ Set the private variable values on instantiation. """

        self.__bypass = bypass
        self.__device_number = device_number
        self.__device_type = device_type
        self.__enrollment_id = enrollment_id
        self.__id = id
        self.__name = name
        self.__partitions = partitions
        self.__preenroll = preenroll
        self.__removable = removable
        self.__renamable = renamable
        self.__subtype = subtype
        self.__warnings = warnings
        self.__zone_type = zone_type

    def __str__(self):
        """ Define how the print() method should print the object. """
        object_type = str(type(self))
        return object_type + ": " + str(self.as_dict())

    def __repr__(self):
        """ Define how the object is represented when output to console. """

        class_name          = type(self).__name__
        bypass              = f"bypass = '{self.bypass}'"
        device_number       = f"device_number = '{self.device_number}'"
        device_type         = f"device_type = {self.device_type}"
        enrollment_id       = f"enrollment_id = '{self.enrollment_id}'"
        id                  = f"id = '{self.id}'"
        name                = f"name = '{self.name}'"
        partitions          = f"partitions = '{self.partitions}'"
        preenroll           = f"preenroll = {self.preenroll}"
        removable           = f"removable = {self.removable}"
        renamable           = f"renamable = {self.renamable}"
        subtype             = f"subtype = {self.subtype}"
        warnings            = f"warnings = {self.warnings}"
        zone_type           = f"zone_type = {self.zone_type}"

        return f"{class_name}({bypass}, {device_number}, {device_type}, {enrollment_id}, {id}, {name}, {partitions}, {preenroll}, {removable}, {renamable}, {subtype}, {warnings}, {zone_type})"

    def as_dict(self):
        """ Return the object properties in a dictionary. """
        return {
            'bypass': self.bypass,
            'device_number': self.device_number,
            'device_type': self.device_type,
            'enrollment_id': self.enrollment_id,
            'id': self.id,
            'name': self.name,
            'partitions': self.partitions,
            'preenroll': self.preenroll,
            'removable': self.removable,
            'renamable': self.renamable,
            'subtype': self.subtype,
            'warnings': self.warnings,
            'zone_type': self.zone_type,
        }

    # Device properties
    @property
    def bypass(self):
        return self.__bypass

    @property
    def device_number(self):
        return self.__device_number

    @property
    def device_type(self):
        return self.__device_type

    @property
    def enrollment_id(self):
        return self.__enrollment_id

    @property
    def id(self):
        return self.__id

    @property
    def name(self):
        return self.__name

    @property
    def partitions(self):
        return self.__partitions

    @property
    def preenroll(self):
        return self.__preenroll

    @property
    def removable(self):
        return self.__removable

    @property
    def renamable(self):
        return self.__renamable

    @property
    def subtype(self):
        return self.__subtype

    @property
    def warnings(self):
        return self.__warnings

    @property
    def zone_type(self):
        return self.__zone_type


class CameraDevice(Device):
    """ Camera device class definition. """
    __location = None
    __soak = None
    __vod = None

    def __init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type, location, soak, vod):
        Device.__init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type)
        self.__location = location
        self.__soak = soak
        self.__vod = vod

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        value_dict = self.as_dict()
        value_dict['location'] = self.__location
        value_dict['soak'] = self.__soak
        value_dict['vod'] = self.__vod
        return object_type + ": " + str(value_dict)


class ContactDevice(Device):
    """ Contact device class definition. """

    __location = None
    __soak = None

    def __init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type, location, soak):
        Device.__init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type)
        self.__location = location
        self.__soak = soak

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        value_dict = self.as_dict()
        value_dict['location'] = self.__location
        value_dict['soak'] = self.__soak
        return object_type + ": " + str(value_dict)

    @property
    def state(self):
        """ Returns the current state of the contact. """
        if self.warnings is None:
            return 'CLOSED'
        for warning in self.warnings:
            if warning['type'] == 'OPENED':
                return 'OPENED'


class GenericDevice(Device):
    """ Smoke device class definition. """
    pass


class GSMDevice(Device):
    """ GSM device class definition. """
    __signal_level = None

    def __init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type, signal_level):
        Device.__init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type)
        self.__signal_level = signal_level

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        value_dict = self.as_dict()
        value_dict['signal_level'] = self.__signal_level
        return object_type + ": " + str(value_dict)


class KeyFobDevice(Device):
    """ KeyFob device class definition. """
    __owner_id = None
    __owner_name = None

    def __init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type, owner_id, owner_name):
        Device.__init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type)
        self.__owner_id = owner_id
        self.__owner_name = owner_name

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        value_dict = self.as_dict()
        value_dict['owner_id'] = self.__owner_id
        value_dict['owner_name'] = self.__owner_name
        return object_type + ": " + str(value_dict)


class PGMDevice(Device):
    """ PGM device class definition. """
    __parent_id = None
    __parent_port = None

    def __init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type, parent_id, parent_port):
        Device.__init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type)
        self.__parent_id = parent_id
        self.__parent_port = parent_port

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        value_dict = self.as_dict()
        value_dict['parent_id'] = self.__parent_id
        value_dict['parent_port'] = self.__parent_port
        return object_type + ": " + str(value_dict)


class SmokeDevice(Device):
    """ Smoke device class definition. """
    __location = None
    __soak = None

    def __init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type, location, soak):
        Device.__init__(self, bypass, device_number, device_type, enrollment_id, id,
                 name, partitions, preenroll, removable, renamable, 
                 subtype, warnings, zone_type)
        self.__location = location
        self.__soak = soak

    def __str__(self):
        """ Define how the print() method should print the object. """

        object_type = str(type(self))
        value_dict = self.as_dict()
        value_dict['location'] = self.__location
        value_dict['soak'] = self.__soak
        return object_type + ": " + str(value_dict)




# define Python user-defined exceptions
class Error(Exception):
    """ Base class for other exceptions """
    pass


class AlreadyGrantedError(Error):
    """ Raised when trying to grant access to a user that has already been granted access. """
    def __init__(self, message="The user has already been granted access."):
        self.message = message
        super().__init__(self.message)


class AlreadyLinkedError(Error):
    """ Raised when trying grant an already linked email address. """
    def __init__(self, message="The email address has already been linked to a user."):
        self.message = message
        super().__init__(self.message)


class AppIDRequiredError(Error):
    """ Raised when an AppID is not provided in the API calls. """
    def __init__(self, message="Connection to the alarm panel failed because the app ID is missing."):
        self.message = message
        super().__init__(self.message)


class EmailRequiredError(Error):
    """ Raised when the email address is missing in the authentication request. """
    def __init__(self, message="Authentication failed because the email address is missing."):
        self.message = message
        super().__init__(self.message)


class LoginTemporaryBlockedError(Error):
    """ Raised when the password is missing in the authentication request. """
    def __init__(self, message="Login is temporary blocked due to too many failed login attempts."):
        self.message = message
        super().__init__(self.message)


class NewPasswordStrengthError(Error):
    """ Raised when the password is missing in the authentication request. """
    def __init__(self, message="New password is not strong enough. Please enter a complex password containing letters, digits and special characters."):
        self.message = message
        super().__init__(self.message)


class NotAllowedError(Error):
    """ Raised when the request is not allowed. """
    def __init__(self, message="The request is not allowed."):
        self.message = message
        super().__init__(self.message)


class PanelNotConnectedError(Error):
    """ Raised when the API server is not connected to the alarm panel. """
    def __init__(self, message="Alarm panel is not connected to the API server."):
        self.message = message
        super().__init__(self.message)


class PanelSerialIncorrectError(Error):
    """ Raised when an incorrect panel serial/ID number was provided in the request. """
    def __init__(self, message="Connection to the alarm panel failed because the panel ID is incorrect."):
        self.message = message
        super().__init__(self.message)


class PanelSerialRequiredError(Error):
    """ Raised when a panel serial/ID number was not provided in the request. """
    def __init__(self, message="Connection to the alarm panel failed because the panel ID is missing."):
        self.message = message
        super().__init__(self.message)


class PasswordRequiredError(Error):
    """ Raised when the password is missing in the authentication request. """
    def __init__(self, message="Request failed because the password is missing."):
        self.message = message
        super().__init__(self.message)


class ResetPasswordCodeIncorrectError(Error):
    """ Raised when the password reset code obtained via email is incorrect. """
    def __init__(self, message="Reset password code is incorrect. Check you email or try to resend the password reset request."):
        self.message = message
        super().__init__(self.message)


class UndefinedBadRequestError(Error):
    """ Raised when an undefined 400 Bad Request error occurs. """
    
    def __init__(self, message="Bad Request. Check the connection details and try again."):
        self.message = message
        super().__init__(self.message)


class UnauthorizedError(Error):
    """ Raised when a 401 Client Error occurs. """

    def __init__(self, message="Unauthorized to access API endpoint."):
        self.message = message
        super().__init__(self.message)


class UndefinedForbiddenError(Error):
    """ Raised when an undefined 403 Client Error occurs. """
    def __init__(self, message="The request is forbidden."):
        self.message = message
        super().__init__(self.message)


class UserAuthRequiredError(Error):
    """ Raised when a forbidden error occurs due to not being authenticated. """
    def __init__(self, message="User authentication required."):
        self.message = message
        super().__init__(self.message)


class UserCodeIncorrectError(Error):
    """ Raised when the user code provided in the request is incorrect. """
    def __init__(self, message="Connection to the alarm panel failed because the user code is incorrect."):
        self.message = message
        super().__init__(self.message)


class UserCodeRequiredError(Error):
    """ Raised when the user code provided in the request was missing. """
    def __init__(self, message="Connection to the alarm panel failed because the user code is missing."):
        self.message = message
        super().__init__(self.message)


class WrongUsernameOrPasswordError(Error):
    """ Raised when the username or password is incorrect. """
    def __init__(self, message="Authentication failed because the wrong username or password was provided."):
        self.message = message
        super().__init__(self.message)


class WrongPanelSerialOrMasterUserCodeError(Error):
    """ Raised when the panel serial or master user code is incorrect. """
    def __init__(self, message="The wrong combination of panel serial and/or master user code was provided."):
        self.message = message
        super().__init__(self.message)


class ConnectionTimeoutError(Error):
    """ Raised when connection to the REST API Server timed out. """
    
    def __init__(self, message="Connection to host timed out."):
        self.message = message
        super().__init__(self.message)


class InvalidPanelIDError(Error):
    """ Raised when the Panel ID is not found in the server. """
    
    def __init__(self, message="Invalid Panel ID."):
        self.message = message
        super().__init__(self.message)


class InvalidUserCodeError(Error):
    """ Raised when the user code is not associated with the alarm system. """
    
    def __init__(self, message="Invalid User Code."):
        self.message = message
        super().__init__(self.message)


class LoginAttemptsLimitReachedError(Error):
    """ Raised when the number of failed login attempts are too many. """
    
    def __init__(self, message="Login attempts limit reached. Please wait a few minutes and then try again."):
        self.message = message
        super().__init__(self.message)


class LoginFailedError(Error):
    """ Raised when the login attempt failed. """
    
    def __init__(self, message="Login attempts failed. Please check the settings and try again. Make sure to use the master code."):
        self.message = message
        super().__init__(self.message)


class NotFoundError(Error):
    """ Raised when an API endpoint is not found. """
    
    def __init__(self, message="The API endpoint was not found on the server."):
        self.message = message
        super().__init__(self.message)


class NotImplementedError(Error):
    """ Raised when a method is called that is not implemented yet. """

    def __init__(self, message="The method called has not yet been implemented."):
        self.message = message
        super().__init__(self.message)


class NotRestAPIError(Error):
    """ Raised when connection to server is not a Rest API. """
    
    def __init__(self, message="Connection to host timed out."):
        self.message = message
        super().__init__(self.message)


class NotSupportedError(Error):
    """ Raised when trying to call a method not supported in the current API version. """
    
    def __init__(self, message="Method is not supported in the selected version of the API."):
        self.message = message
        super().__init__(self.message)


class SessionTokenError(Error):
    """ Raised when not authenticated with the REST API. """
    
    def __init__(self, message="Session token not found. Please log in prior to this call."):
        self.message = message
        super().__init__(self.message)


class UnsupportedRestAPIVersionError(Error):
    """ Raised when a version of the REST API is unsupported. """
    
    def __init__(self, message="Unsupported REST API version."):
        self.message = message
        super().__init__(self.message)
